% 
%   MLCV Winter 2015
%   Homework Assignment 03 - Hidden Markov Models
%
%   Exercise 1d) Viterbi algorithm
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Initializations
Map = [ 'G', 'R', 'B'; 
        'R', 'B', 'G'; 
        'B', 'G', 'R'];

numStates = numel(Map);
states = cell(numStates,1);
for x=1:3
    for y=1:3
        states{ (x-1)*3 + y} = [x y];
    end
end

stateLabels = cellfun(@(s) sprintf('(%d,%d)', s(1), s(2)), states, 'UniformOutput', 0);
observationLabels = {'R' 'G' 'B'}';

isValidState = @(s) any(cellfun(@(x) isequal(x, s), states, 'UniformOutput', 1));
getStateIdx = @(s) find(cellfun(@(x) isequal(x,s), states, 'UniformOutput', 1));
getObservationIdx = @(o) find(cellfun(@(x) isequal(o,x), observationLabels, 'UniformOutput', 1)); 


% Define state transition matrix
P_trans = zeros(numStates); 
moves = {[0 1]; [0 -1]; [-1 0]; [1 0]}; % up, down, left, right
p_move = 1 / numel(moves);

for i=1:numStates
    curState = states{i};   
    % Try to move up, down, left, right
    for m=1:length(moves)
        newState = curState + moves{m};
        if ~isValidState(newState) % out of bounds
            newState = curState;
        end
        j = getStateIdx(newState);      
        P_trans(i, j) = P_trans(i, j) + p_move;
    end   
end

% Define observation model p(z|x) from first exercise 1 
% Each column is the probability to sense color z, if given color x
P_zx = [0.8 0.1 0.1; 
        0.1 0.6 0.2; 
        0.1 0.3 0.7];

ObservationModel = table(P_zx(:,1), P_zx(:,2), P_zx(:,3), ... 
                    'VariableNames', observationLabels, ... 
                    'RowNames', observationLabels)
                
P_sense = zeros(length(observationLabels), numStates);
for i=1:numStates
    stateColor = Map(states{i}(1), states{i}(2));
    idx = getObservationIdx(stateColor);
    P_sense(:,i) = P_zx(:,idx);
end
        
% Define initial probabilities p(x0)
P_init = ones(numStates,1) / numStates;


% Observed sequence
observations = {'R', 'G', 'G'};
observSequence = cellfun(@(x) getObservationIdx(x), observations);

[ probVit, pathVit ] = viterbi_fun(observSequence, P_init, P_trans, P_sense, ... 
                                     stateLabels, observationLabels)
